//++
// eeload.c
//
// Copyright (C) 2000 by Robert Armstrong.
//
//   This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
//   You should have received a copy of the GNU General Public License along
// with this program; if not, visit the website of the Free Software
// Foundation, Inc., www.gnu.org.
//
//
//   This program downloads the EPROM emulator from ROM images stored in PC
// files.  The emulator contains two 32Kx8 SRAMs that can be used to emulate
// either one or two 32K 27C256 type EPROMs.  However, the emulator board has
// only a single set of multiplexors to switch the SRAM address inputs between
// the target under test and the internal counters used to download the
// emulator. This means that when we want to emulate two EPROMs they cannot be
// completely independent and must have all address inputs in common, however
// they can have different chip selects and output enables. Even with this
// restriction, the emulator can still be used to emulate two EPROMs in
// parallel for 16 bit wide memory on a 16 bit processor, or it can emulate
// two EPROMs used for different address ranges (i.e. different chip selects)
// in an 8 bit architecture.
//
// Usage:
//   eeload [-pxxx] [-t] [-v] [-d] [-oxxxx] file-1 [-oxxxx] [file-2]
//
// Options:
//   -pxxx      - select the LPT port for the emulator
//   -t         - execute an emulator test diagnostic
//   -v         - read back and verify emulator memory
//   -d         - dump (don't load!) emulator memory
//   -oxxxx     - set ROM offset
//
//   Image file-1 is loaded into what is usually the low order byte EPROM and
// image file-2 into the high order byte.  However this assignment is somewhat
// arbitrary and it really depends on how you connect the emulator cables to
// the target under test.  The second image file may be omitted, as it would
// be if we were emulating only a single EPROM.  The extension of the image
// file determines its data format, and may be either .HEX for Intel HEX files
// or .BIN for pure binary images.
//
//   The -o option is used to set the base address or offset of the ROM.  For
// example, if a 32K EPROM were mapped into addresses 0x8000 to 0xFFFF on the
// target system, then the Intel HEX file would presumably contain data with
// load addresses in the range 0x8000 to 0xFFFF.  In this case you would give
// the -o0x8000 option to relocate the ROM data so that address 0x8000 in the
// HEX file would load into ROM address 0x0000, and so on.  If only one -o
// is specified then the same offset is used for both ROMs, or two -o options
// may be used to specify different offsets for each ROM.  Note that -o is
// only useful with Intel HEX format input - binary ROM image files do not
// contain address information!
//
// REVISION HISTORY
// 10-Feb-00    RLA             New file.
//--
#include <stdio.h>              // printf(), scanf(), etc...
#include <stdlib.h>             // malloc(), free(), exit(), ...
#include <malloc.h>                     // _halloc(), _hfree(), ...
#include <string.h>             // strcmp(), etc...
#include "romtools.h"           // standard ROM utilities (e.g. LoadHex(), etc)
#include "eeload.h"             // declarations for this project alone

// Global variables...
char szFile[2][_MAX_PATH];      // file name of both image files
long   lOffset[2];                      // offset for both ROMs
UINT   nFiles;                          // number of files specified (0, 1 or 2)
BOOL   fTest;                           // TRUE to execute the full diagnostic
BOOL   fVerify;                         // TRUE to read back and verify memory
BOOL   fDump;                           // TRUE to dump emulator memory (not load!)
UINT   nPort;                           // I/O address of LPT port (e.g. 0x378)
UINT   nROMSize;                        // size of emulator memory (words!)
HPBYTE hpbImage[2];                     // pointer to the ROM image data


//++
//   This function parses the command line and initializes all the global
// variables accordingly.  It's tedious, but fairly brainless work.  If there 
// are any errors in the command line, then it simply prints an error message
// and exits - there is no error return from this function!
//--
PRIVATE void ParseCommand (int argc, char *argv[])
{
  int nArg;  char *psz;

  // First, set all the defaults...
  nFiles = 0;  nPort = LPT1;
  fTest = fVerify = fDump = FALSE;  
  szFile[0][0] = szFile[1][0] = '\0';
  lOffset[0] = lOffset[1] = 0;
  
  // If there are no arguments, then just print the help and exit...
  if (argc == 1) {
    fprintf(stderr, "Usage:\n");
    fprintf(stderr,"\teeload [-pxxx] [-t] [-v] [-d] [-oxxxx] file-1 [-oxxxx] [file-2]\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"Options:\n");
    fprintf(stderr,"\t-pxxx\t- select the LPT port for the emulator\n");
    fprintf(stderr,"\t-t\t- execute an emulator test diagnostic\n");
        fprintf(stderr,"\t-v\t- read back and verify emulator memory\n");
        fprintf(stderr,"\t-d\t- dump (don't load!) emulator memory\n");
    fprintf(stderr,"\t-oxxxx\t- set ROM offset\n");
    exit(EXIT_SUCCESS);
  }

  for (nArg = 1;  nArg < argc;  ++nArg) {
    // If it doesn't start with a "-" character, then it must be a file name.
    if (argv[nArg][0] != '-') {
      if (++nFiles > 2)
        FAIL1("EELOAD", "too many files specified: \"%s\"", argv[nArg]);
      strcpy(szFile[nFiles-1], argv[nArg]);  continue;
    }
    
    // Handle the -t (test) option...
    if (strcmp(argv[nArg], "-t") == 0) {
      fTest = TRUE;  continue;
    }
    
    // Handle the -v (verify) option...
    if (strcmp(argv[nArg], "-v") == 0) {
      fVerify = TRUE;  continue;
    }
    
    // Handle the -d (dump) option...
    if (strcmp(argv[nArg], "-d") == 0) {
      fDump = TRUE;  continue;
    }
    
    // Handle the -p (port) option...
    if (strncmp(argv[nArg], "-p", 2) == 0) {
      nPort = (UINT) strtoul(argv[nArg]+2, &psz, 16);
      if ((*psz != '\0')  ||  (nPort == 0))
        FAIL1("EELOAD", "illegal port number: \"%s\"", argv[nArg]);
      //   If the port is 1, 2 or 3 then it's the number of the LPT port, but
      // if it's anything else, then it's the absolute port address...
      if (nPort == 1) nPort = LPT1;
      if (nPort == 2) nPort = LPT2;
      if (nPort == 3) nPort = LPT3;
      continue;
    }
    
    // Handle the -o (offset) option...
    if (strncmp(argv[nArg], "-o", 2) == 0) {
      //   Note that -o comes _before_ the file name, so nFile hasn't been
      // incremented yet!
      if (nFiles >= 2) FAIL("EELOAD", "use -o before a file name");
      lOffset[nFiles] = strtoul(argv[nArg]+2, &psz, 16);
      if (*psz != '\0') FAIL1("EELOAD", "illegal offset: \"%s\"", argv[nArg]);
      continue;
    }
    
    // Otherwise it's an illegal option...
    FAIL1("EELOAD", "unknown option - \"%s\"\n", argv[nArg]);
  }

  // Check the options given to make sure that the combination is legal...
  if ((nFiles == 0)  &&  (!fTest || fDump))
    FAIL("EELOAD", "at least one file name must be specified");
  if (fDump && fTest)
    FAIL("EELOAD", "-d and -t cannot be used together");
}


//++
//   This routine is an exhaustive emulator diagnostic that tests all memory
// locations, address counters, and data buffers.  It's a little slow, so we
// use it only when explicitly requested by the -t option...
//--
PRIVATE BOOL TestEmulator (void)
{
  UINT n, i;

  fprintf(stderr, "EELOAD: memory test ... ");  

  //   Pass one is a simple addressing test - to each pair of bytes we write
  // its 16 bit address and verify it.  This tests that each location can be
  // unique addressed (well sort of - read on).
  for (n = 0;  n < nROMSize;  ++n)
    {hpbImage[0][n] = LOBYTE(n);  hpbImage[1][n] = HIBYTE(n);}
  LoadEmulator(nPort, hpbImage[0], hpbImage[1], nROMSize);
  if (!VerifyEmulator(nPort, hpbImage[0], hpbImage[1], nROMSize)) return FALSE;
  fprintf(stderr, "0");

  //   That was all fine, but it didn't actually test that all the address bits
  // are correct.  Since the SRAMs are only 8 bits, their contents repeat every
  // 256 bytes.  Thus the upper address bits for the low order byte or the lower
  // address bits for the high order byte could have been wired wrong and we'd
  // never notice.  To guard against that possibility, we do the same test all
  // over again with the bytes swapped...
  for (n = 0;  n < nROMSize;  ++n)
    {hpbImage[0][n] = HIBYTE(n);  hpbImage[1][n] = LOBYTE(n);}
  LoadEmulator(nPort, hpbImage[0], hpbImage[1], nROMSize);
  if (!VerifyEmulator(nPort, hpbImage[0], hpbImage[1], nROMSize)) return FALSE;
  fprintf(stderr, "1");

  //   Those tests are probably enough for real life, but just to be completely
  // thorough we'll now do both a sliding one and a sliding zero test on both
  // bytes.  The idea is to individually test each bit of all locations...
  for (i = 0;  i < 8;  ++i) {
    for (n = 0;  n < nROMSize;  ++n)
      {hpbImage[0][n] = 1 << i;  hpbImage[1][n] = ~(1 << i);}
    LoadEmulator(nPort, hpbImage[0], hpbImage[1], nROMSize);
    if (!VerifyEmulator(nPort, hpbImage[0], hpbImage[1], nROMSize)) return FALSE;
    for (n = 0;  n < nROMSize;  ++n)
      {hpbImage[0][n] = ~(1 << i);  hpbImage[1][n] = 1 << i;}
    LoadEmulator(nPort, hpbImage[0], hpbImage[1], nROMSize);
    if (!VerifyEmulator(nPort, hpbImage[0], hpbImage[1], nROMSize)) return FALSE;
    fprintf(stderr, "%c", 'A'+i);
  }

  // And _that's_ really enough...      
  fprintf(stderr," ... Passed!\n");
}


//++
// The main program...
//--                                                        
int main (int argc, char *argv[])
{
  ParseCommand(argc, argv);

  //   Figure out the amount of memory installed in the emulator and allocate
  // two blocks of memory to hold the ROM images.  This is also where we find
  // out whether the emulator hardware is really connected and working or not.
  // If it isn't then the SizeEmulator() function will fail!
  nROMSize = SizeEmulator(nPort);
  if (nROMSize == 0)
    FAIL1("EELOAD", "no emulator detected on port 0x%03X", nPort);
  fprintf(stderr, "EELOAD: %dKW emulator found on port 0x%03X\n", nROMSize>>10, nPort);
  hpbImage[0] = _halloc(nROMSize, sizeof(BYTE));
  hpbImage[1] = _halloc(nROMSize, sizeof(BYTE));
  
  //   The emulator diagnostic destroys the current memory contents, so if we
  // want to dump the emulator contents, then we'd better do it first!
  if (fDump) {
    fprintf(stderr, "EELOAD: uploading ... ");
    DumpEmulator(nPort, hpbImage[0], hpbImage[1], nROMSize);
    fprintf(stderr, " %u words done\n", nROMSize);
    if (!DumpHexOrBinary(szFile[0], hpbImage[0], 0, nROMSize)) return EXIT_FAILURE;
    if (nFiles == 2)
      if (!DumpHexOrBinary(szFile[1], hpbImage[1], 0, nROMSize)) return EXIT_FAILURE;
  }

  //   If the user wants a full memory diagnostic (rather than a simple read
  // back and verify) then do it before loading the data...    
  if (fTest) {
    if (!TestEmulator()) return EXIT_FAILURE;
    if (nFiles == 0) return EXIT_SUCCESS;
  }

  // And lastly we can read the data files and download them to the emulator...
  if (!fDump) {
    if (!LoadHexOrBinary(szFile[0], hpbImage[0], lOffset[0], nROMSize)) return EXIT_FAILURE;
    if (nFiles == 2)
      if (!LoadHexOrBinary(szFile[1], hpbImage[1], lOffset[1], nROMSize)) return EXIT_FAILURE;
    //   There's a small trick here - if only one data file was specified, then
    // we load the same data into both bytes of the emulator.  That way we don't
    // have to worry about getting the cables swapped!
    fprintf(stderr, "EELOAD: downloading ... ");
    LoadEmulator(nPort, hpbImage[0], (nFiles==2 ? hpbImage[1] : hpbImage[0]), nROMSize);
    if (fVerify) {
      fprintf(stderr, "verifying ... ");
      if (!VerifyEmulator(nPort, hpbImage[0], (nFiles==2 ? hpbImage[1] : hpbImage[0]), nROMSize)) return EXIT_FAILURE;
    }
    fprintf(stderr, " %u words done\n", nROMSize);
  }      

  // Release the memory buffers and we're all done...
  _hfree(hpbImage[0]);  _hfree(hpbImage[1]);  
  return EXIT_SUCCESS;
}
                                                        